/*********************************************************************************
 *
 * cfa-10004.c
 *
 * Test code for the CFA-10004 carrier board with the Crystalfontz
 * CFAO4265A 1" LCD.
 *
 * The CFAO4265A is based on a Sitronix ST7565P controller which has
 * the serial SPI interface enabled. This is a read-only interface to
 * the controller. The controller interface includes the following
 * signals:
 *
 *   SPI_MOSI - Master out, slave in, serial data to controller
 *   SPI_SCK  - Clock to the controller, clocks in serial data
 *   LCD_CS   - Active low chip select
 *   LCD_A0   - 0 = command, 1 = data
 *   LCD_RST  - Active low reset to controller
 *
 * The CFAO4265A has a white LED backlight that is driven by a LT3491
 * White LED Driver IC. The control voltage to the LT3491, which sets
 * the backlight brightness, is driven by the output of a LTC2630-8
 * DAC. The LT3491 input control voltage range is 0V (backlight off)
 * to 1.5V (backlight full on). The DAC is initialized to use VCC as
 * its reference voltage resulting in a full output swing of 0V to
 * 3.3V. The DAC output voltage to the LT3491 control input is
 * resistor divided by 2 resulting in a control voltage range of 0V to
 * 1.65V. The LTC2630-8 interface includes the following signals:
 *
 *   SPI_MOSI - Master out, slave in, serial data to DAC
 *   SPI_SCK  - Clock to the controller, clocks in serial data
 *   DAC_CS   - Active low chip select
 *
 * In this application, the CFA-10004 carrier board is connected to a
 * Wizbang Designs 10011-644V-ASSY Development Board. The 10011
 * development board is based on the ATmega644 microcontroller with a
 * 3.6868MHz oscillator.
 *
 *
 *
 *********************************************************************************/

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "cfa-10004.h"
#include "CFAO4265_logo.h"

extern uint8_t cfao4265_logo[];


void spi_init(void)
{
     /* Initialize SPI as master sck = fck/16, data clocked on rising
      * edge. The SPI clock may be set as high as 20MHz for the
      * ST7565P LCD controller on the CFAO4265.
      */
     SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR0);

}

void spi_master_send_byte(uint8_t data)
{
     /* Start data transmission
      */
     SPDR = data;

     /* Wait for transmission complete
      */
     while (!(SPSR & (1 << SPIF)));
}


void lcd_reset(void)
{
     /* Set the LCD_RST low for 10 uS and then high again and return
      * after 10 uS. According to the ST7565P datasheet only 1 uS is
      * needed at VDD = 3.3VDC but it won't hurt to be conservative.
      */
     LCD_PORT &= ~(1 << LCD_RST);
     _delay_us(10);
     LCD_PORT |= (1 << LCD_RST);
     _delay_us(10);
}


void lcd_cmd_send(uint8_t data)
{
     /* Set LCD_CS low and LCD_A0 low to indicate command to LCD Send
      * command byte to LCD
      */
     LCD_PORT &= ~((1 << LCD_CS) | (1 << LCD_A0));
     spi_master_send_byte(data);

     /* set LCD_CS high and LCD_A0 high again
      */
     LCD_PORT |= ((1 << LCD_CS) | (1 << LCD_A0));

}


void lcd_data_send(uint8_t data)
{
     /* Set LCD_CS low and LCD_A0 high to indicate data to LCD Send
      * data byte to LCD
      */
     LCD_PORT &= ~(1 << LCD_CS);
     LCD_PORT |= (1 << LCD_A0);
     spi_master_send_byte(data);

     /* set LCD_CS high and LCD_A0 high again
      */
     LCD_PORT |= ((1 << LCD_CS) | (1 << LCD_A0));
}


void lcd_init(void)
{
     lcd_reset();
     lcd_bias_set(LCD_BIAS_1_7);
     lcd_adc_select(NORMAL);
     lcd_common_output_mode_select(NORMAL);
     lcd_vo_regulator_resistor_set(LCD_RATIO_5_0);
     lcd_electronic_volume_set(36);
     lcd_booster_ratio_set(LCD_BOOSTER_4X);
     lcd_power_controller_set(BOOST_ON, VR_ON, VF_ON);
     lcd_start_line_set(0);
     lcd_clear();
}


void lcd_display_on(void)
{
     lcd_cmd_send(0b10101111);
}

void lcd_display_off(void)
{
     lcd_cmd_send(0b10101110);
}

void lcd_start_line_set(uint8_t line)
{
     lcd_cmd_send(0b01000000 | (line & 0b111111));
}

void lcd_page_address_set(uint8_t page)
{
     lcd_cmd_send(0b10110000 | (page & 0b1111));
}

void lcd_column_address_set(uint8_t column)
{
     lcd_cmd_send(0b00010000 | (column >> 4));
     lcd_cmd_send(0b00000000 | (column & 0b1111));
}

void lcd_adc_select(uint8_t direction)
{
     lcd_cmd_send(0b10100000 | (direction & 0b1));
}

void lcd_display_reverse(uint8_t reverse)
{
     lcd_cmd_send(0b101000110 | (reverse & 0b1));
}

void lcd_all_points(uint8_t on)
{
     lcd_cmd_send(0b10100100 | (on & 0b1));
}

void lcd_bias_set(uint8_t bias)
{
     lcd_cmd_send(0b10100010 | (bias & 0b1));
}

void lcd_cmd_reset(void)
{
     lcd_cmd_send(0b11100010);
}

void lcd_common_output_mode_select(uint8_t reverse)
{
     lcd_cmd_send(0b11000000 | ((reverse & 0b1) << 3));
}

void lcd_power_controller_set(uint8_t boost, uint8_t vreg, uint8_t vfollower)
{
     lcd_cmd_send(0b00101000 | ((boost & 0b1) << 2)
                  | ((vreg & 0b1) << 1)
                  | (vfollower & 0b1));
}

void lcd_electronic_volume_set(uint8_t contrast)
{
     lcd_cmd_send(0b10000001);
     lcd_cmd_send(contrast & 0b00111111);
}

void lcd_vo_regulator_resistor_set(uint8_t ratio)
{
     lcd_cmd_send(0b00100000 | (ratio & 0b111));
}

void lcd_booster_ratio_set(uint8_t ratio)
{
     lcd_cmd_send(0b11111000);
     lcd_cmd_send(ratio & 0b11);
}

void lcd_clear(void)
{
     uint8_t page, col;

     for (page = 0; page < 9; ++page)
     {
          lcd_page_address_set(page);
          lcd_column_address_set(0);
          for (col = 0; col < 42; ++col)
          {
               lcd_data_send(0);
          }
     }
}

void lcd_fill(uint8_t data, uint16_t delay_ms_col, uint16_t delay_ms_page)
{
     uint8_t page, col;

     for (page = 0; page < 9; ++page)
     {
          lcd_page_address_set(page);
          lcd_column_address_set(0);
          for (col = 0; col < 42; ++col)
          {
               lcd_data_send(data);
               _delay_ms(delay_ms_col);

          }
          _delay_ms(delay_ms_page);
     }
}

void lcd_backlight_set(uint8_t light)
{
     dac_cmd_send(DAC_REG_WRITE_UPDATE, light);
}


void lcd_backlight_cycle(void)
{
     uint8_t light_val;

     /* Decrease backlight from full brightness to almost off
      */
     for (light_val = BACKLIGHT_FULL_ON; light_val > BACKLIGHT_FULL_OFF; --light_val)
     {
          lcd_backlight_set(light_val);
          _delay_ms(40);
     }

     /* Leave backlight off for 1S
	  */
     delay_100ms(100);

     /* Increase backlight from off to almost full brightness
      */
     for (light_val = BACKLIGHT_FULL_OFF; light_val < BACKLIGHT_FULL_ON; ++light_val)
     {
          lcd_backlight_set(light_val);
          _delay_ms(40);
     }
}


void dac_init(void)
{
     /* Initialize the LTC2630-8 DAC. Set its reference to VCC
      * (+3.3VDC)
      */
     dac_cmd_send(DAC_POWER_DOWN, NULL_DATA);
     dac_cmd_send(DAC_VCC_REF, NULL_DATA);

}

void dac_cmd_send(dac_cmd_t cmd, uint8_t data)
{
     /* Set DAC_CS low. Send DAC command, data byte followed by don't
      * care byte to DAC
      */
     DAC_PORT &= ~(1 << DAC_CS);
     spi_master_send_byte(cmd);
     spi_master_send_byte(data);
     spi_master_send_byte(NULL_DATA);

     /* Set DAC_CS high
      */
     LCD_PORT |= (1 << DAC_CS);
}


void gpio_init(void)
{
     /* Initialize PORT A
      *
      * bits [7:2] input,  pull-up  (UNUSED)
      * bit  [1]   output, low      (LED2 OFF)
      * bit  [0]   output, low      (LED1 OFF)
      */
     DDRA  = 0b00000011;
     PORTA = 0b11111100;

     /* Initialize PORT B
      *
      * bit  [7]   output, low      (SPI_SCK)
      * bit  [6]   input,  pull-up  (SPI_MISO)
      * bit  [5]   output, low      (SPI_MOSI)
      * bit  [4]   output, high     (LCD_CS)
      * bit  [3]   output, high     (LCD_A0)
      * bit  [2]   output, high     (LCD_RST)
      * bit  [1]   output, high     (DAC_CS)
      * bit  [0]   output, high     (SD_CS)
      */
     DDRB  = 0b10111111;
     PORTB = 0b01011111;

     /* Initialize PORT C - Not Used
      */
     DDRC  = 0b00000000;
     PORTC = 0b11111111;

     /* Initialize PORT D - Not Used
      */
     DDRD  = 0b00000000;
     PORTD = 0b11111111;


}

void delay_100ms(uint8_t tenth_seconds)
{
     do
     {
          _delay_ms(100);
     } while (tenth_seconds-- != 0);
}


int main(void)
{
     uint8_t page, col, data1, data2;
     uint16_t i = 0;


     /* Initialize the hardware
      */
     gpio_init();
     spi_init();
     dac_init();
     lcd_init();

     /* Turn on the display and its backlight
      */
     lcd_display_on();
     lcd_backlight_set(BACKLIGHT_FULL_OFF);


     /* Loop through the demo forever
      */
     for (;;)
     {

          /* Display the Crystalfontz CFAO4265 bitmap image.
           *
           * The bitmap2lcd utility generated the data with the image
           * shifted up 7 pixels from the bottom of the LCD. This is
           * handled prior to display of each byte by shifting the
           * byte from the current page down 7 pixels and ORing it
           * with the byte from the next page that is shifted up 1
           * pixel.
           */
          for (i = 0, page = 0; page < 9; ++page)
          {
               lcd_page_address_set(page);
               lcd_column_address_set(0);
               for (col = 0; col < 42; ++col)
               {
                    data1 = pgm_read_byte(&(cfao4265_logo[i]));
                    data2 = pgm_read_byte(&(cfao4265_logo[i+42]));
                    lcd_data_send((data1 >> 7) | (data2 << 1));
                    ++i;
               }
          }

          /* Delay 4S, cycle the backlight
           * Delay 1.5S then clear the LCD
           */
          delay_100ms(40);
          lcd_backlight_cycle();
          delay_100ms(15);
          lcd_clear();


          /* Display a fill pattern then clear
           */
          lcd_fill(0x3c, 3, 0);
          lcd_fill(0x00, 3, 0);

          /* Display a fill pattern then clear
           */
          lcd_fill(0x03c, 0, 255);
          lcd_fill(0x00, 0, 255);
     }
}
